# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
from datetime import timedelta
from datetime import timezone

import pytest

from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.repo import result_type
from ooresults.repo.result_type import ResultStatus
from ooresults.repo import start_type


S1 = datetime.datetime(2021, 8, 19, 18, 43, 33, tzinfo=timezone(timedelta(hours=2)))
S2 = datetime.datetime(2021, 8, 18, 17, 55, 00, tzinfo=timezone(timedelta(hours=2)))
S3 = datetime.datetime(2021, 8, 17, 10, 12, 14, tzinfo=timezone(timedelta(hours=2)))


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def event_1_id(db):
    return db.add_event(
        name="event 1",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def event_2_id(db):
    return db.add_event(
        name="event 2",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def class_1_id(db, event_2_id):
    return db.add_class(
        event_id=event_2_id,
        name="Class 1",
        short_name=None,
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def class_2_id(db, event_2_id):
    return db.add_class(
        event_id=event_2_id,
        name="Class 2",
        short_name=None,
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def club_id(db):
    return db.add_club(
        name="OL Bundestag",
    )


@pytest.fixture
def competitor_1_id(db, club_id):
    return db.add_competitor(
        first_name="Angela",
        last_name="Merkel",
        club_id=club_id,
        gender="",
        year=None,
        chip="1234567",
    )


@pytest.fixture
def competitor_2_id(db):
    return db.add_competitor(
        first_name="Jogi",
        last_name="Löw",
        club_id=None,
        gender="",
        year=None,
        chip="",
    )


@pytest.fixture
def entry_1_id(db, event_1_id, class_1_id, club_id):
    return db.add_entry(
        event_id=event_1_id,
        competitor_id=None,
        first_name="Robert",
        last_name="Lewandowski",
        gender="",
        year=None,
        class_id=class_1_id,
        club_id=club_id,
        not_competing=False,
        chip="9999999",
        fields={0: "ab", 1: "cd"},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )


@pytest.fixture
def entry_2_id(db, event_2_id, class_2_id, club_id):
    return db.add_entry(
        event_id=event_2_id,
        competitor_id=None,
        first_name="Angela",
        last_name="Merkel",
        gender="F",
        year=1957,
        class_id=class_2_id,
        club_id=club_id,
        not_competing=True,
        chip="9999999",
        fields={0: "x"},
        status=ResultStatus.DID_NOT_START,
        start_time=S1,
    )


@pytest.fixture
def entry_3_id(db, event_2_id, class_2_id):
    return db.add_entry(
        event_id=event_2_id,
        competitor_id=None,
        first_name="Jogi",
        last_name="Löw",
        gender="M",
        year=None,
        class_id=class_2_id,
        club_id=None,
        not_competing=False,
        chip="",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )


def test_get_entries(
    db, event_2_id, entry_1_id, entry_2_id, entry_3_id, club_id, class_2_id
):
    data = db.get_entries(event_id=event_2_id)
    print(data[0])
    assert len(data) == 2

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == ""
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.INACTIVE
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_2_id
    assert data[0]["class_"] == "Class 2"

    assert data[1]["id"] == entry_2_id
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == "F"
    assert data[1]["year"] == 1957
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "9999999"
    assert data[1]["fields"] == {0: "x"}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_2_id
    assert data[1]["class_"] == "Class 2"


def test_get_first_added_entry(
    db, event_2_id, club_id, class_2_id, entry_1_id, entry_2_id, entry_3_id
):
    competitors = list(db.get_competitors())
    competitor = [
        c for c in competitors if c.first_name == "Angela" and c.last_name == "Merkel"
    ][0]

    assert competitor["first_name"] == "Angela"
    assert competitor["last_name"] == "Merkel"
    assert competitor["gender"] == "F"
    assert competitor["year"] == 1957
    assert competitor["chip"] == "9999999"
    assert competitor["club_name"] == "OL Bundestag"
    assert competitor["club_id"] == club_id

    data = list(db.get_entry(id=entry_2_id))
    assert len(data) == 1

    assert data[0]["id"] == entry_2_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["competitor_id"] == competitor.id
    assert data[0]["first_name"] == "Angela"
    assert data[0]["last_name"] == "Merkel"
    assert data[0]["gender"] == "F"
    assert data[0]["year"] == 1957
    assert data[0]["not_competing"] is True
    assert data[0]["chip"] == "9999999"
    assert data[0]["fields"] == {0: "x"}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[0]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[0]["club"] == "OL Bundestag"
    assert data[0]["club_id"] == club_id
    assert data[0]["class_"] == "Class 2"
    assert data[0]["class_id"] == class_2_id


def test_get_last_added_entry(
    db, event_2_id, club_id, class_2_id, entry_1_id, entry_2_id, entry_3_id
):
    competitors = list(db.get_competitors())
    competitor = [
        c for c in competitors if c.first_name == "Jogi" and c.last_name == "Löw"
    ][0]

    assert competitor["first_name"] == "Jogi"
    assert competitor["last_name"] == "Löw"
    assert competitor["gender"] == "M"
    assert competitor["year"] is None
    assert competitor["chip"] == ""
    assert competitor["club_name"] is None
    assert competitor["club_id"] is None

    data = list(db.get_entry(id=entry_3_id))
    assert len(data) == 1

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["competitor_id"] == competitor.id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == ""
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.INACTIVE
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club"] is None
    assert data[0]["club_id"] is None
    assert data[0]["class_"] == "Class 2"
    assert data[0]["class_id"] == class_2_id


def test_update_first_added_entry(
    db, event_2_id, class_1_id, class_2_id, entry_2_id, entry_3_id
):
    db.update_entry(
        id=entry_2_id,
        first_name="angela",
        last_name="merkel",
        gender="",
        year=None,
        class_id=class_1_id,
        club_id=None,
        not_competing=False,
        chip="7788",
        fields={},
        status=ResultStatus.DISQUALIFIED,
        start_time=S2,
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] is None
    assert data[0]["chip"] == ""
    assert data[0]["not_competing"] is False
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.INACTIVE
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_2_id
    assert data[0]["class_"] == "Class 2"

    assert data[1]["id"] == entry_2_id
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "angela"
    assert data[1]["last_name"] == "merkel"
    assert data[1]["gender"] == ""
    assert data[1]["year"] is None
    assert data[1]["not_competing"] is False
    assert data[1]["chip"] == "7788"
    assert data[1]["fields"] == {}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DISQUALIFIED
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S2)
    assert data[1]["club_id"] is None
    assert data[1]["club"] is None
    assert data[1]["class_id"] == class_1_id
    assert data[1]["class_"] == "Class 1"


def test_update_result_first_added_entry(
    db, event_2_id, class_1_id, class_2_id, club_id, entry_2_id, entry_3_id
):
    db.update_entry_result(
        id=entry_2_id,
        chip="7788",
        start_time=S2,
        result=result_type.PersonRaceResult(status=ResultStatus.DISQUALIFIED),
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == ""
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.INACTIVE
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_2_id
    assert data[0]["class_"] == "Class 2"

    assert data[1]["id"] == entry_2_id
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == "F"
    assert data[1]["year"] == 1957
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "7788"
    assert data[1]["fields"] == {0: "x"}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DISQUALIFIED
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S2)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_2_id
    assert data[1]["class_"] == "Class 2"


def test_update_last_added_entry(
    db, event_2_id, class_1_id, class_2_id, club_id, entry_2_id, entry_3_id
):
    db.update_entry(
        id=entry_3_id,
        first_name="jogi",
        last_name="Loew",
        gender="M",
        year=1960,
        class_id=class_1_id,
        club_id=club_id,
        not_competing=False,
        chip="4455",
        fields={},
        status=ResultStatus.DID_NOT_START,
        start_time=S3,
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "jogi"
    assert data[0]["last_name"] == "Loew"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1960
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "4455"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[0]["start"] == start_type.PersonRaceStart(start_time=S3)
    assert data[0]["club_id"] == club_id
    assert data[0]["club"] == "OL Bundestag"
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"

    assert data[1]["id"] == entry_2_id
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == "F"
    assert data[1]["year"] == 1957
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "9999999"
    assert data[1]["fields"] == {0: "x"}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_2_id
    assert data[1]["class_"] == "Class 2"


def test_update_result_last_added_entry(
    db, event_2_id, class_1_id, class_2_id, club_id, entry_2_id, entry_3_id
):
    db.update_entry_result(
        id=entry_3_id,
        chip="4455",
        result=result_type.PersonRaceResult(status=ResultStatus.DID_NOT_START),
        start_time=S3,
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] == entry_3_id
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "4455"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[0]["start"] == start_type.PersonRaceStart(start_time=S3)
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_2_id
    assert data[0]["class_"] == "Class 2"

    assert data[1]["id"] == entry_2_id
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == "F"
    assert data[1]["year"] == 1957
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "9999999"
    assert data[1]["fields"] == {0: "x"}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_2_id
    assert data[1]["class_"] == "Class 2"


def test_add_existing_competitor_but_do_not_update_competitors_chip_and_club_if_alread_defined(
    db, event_1_id, class_1_id, competitor_1_id, club_id
):
    db.add_entry(
        event_id=event_1_id,
        competitor_id=competitor_1_id,
        first_name="Angela",
        last_name="Merkel",
        gender="W",
        year=None,
        class_id=class_1_id,
        club_id=club_id,
        not_competing=False,
        chip="999",
        fields={},
        status=result_type.ResultStatus.INACTIVE,
        start_time=None,
    )
    data = list(db.get_entries(event_id=event_1_id))
    assert len(data) == 1

    assert data[0]["id"] is not None
    assert data[0]["event_id"] == event_1_id
    assert data[0]["competitor_id"] == competitor_1_id
    assert data[0]["first_name"] == "Angela"
    assert data[0]["last_name"] == "Merkel"
    assert data[0]["gender"] == "W"
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "999"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult()
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] == club_id
    assert data[0]["club"] == "OL Bundestag"
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"
    assert data[0]["result"] == result_type.PersonRaceResult()
    assert data[0]["start"] == start_type.PersonRaceStart()

    data = list(db.get_competitor(id=competitor_1_id))
    assert len(data) == 1

    assert data[0]["id"] == competitor_1_id
    assert data[0]["first_name"] == "Angela"
    assert data[0]["last_name"] == "Merkel"
    assert data[0]["gender"] == "W"
    assert data[0]["year"] is None
    assert data[0]["chip"] == "1234567"
    assert data[0]["club_id"] == club_id


def test_add_existing_competitor_and_update_competitors_chip_and_club_if_undefined(
    db, event_1_id, class_1_id, competitor_2_id, club_id
):
    db.add_entry(
        event_id=event_1_id,
        competitor_id=competitor_2_id,
        first_name="Yogi",
        last_name="Löw",
        gender="M",
        year=1975,
        class_id=class_1_id,
        club_id=club_id,
        not_competing=False,
        chip="999",
        fields={},
        status=result_type.ResultStatus.INACTIVE,
        start_time=None,
    )
    data = list(db.get_entries(event_id=event_1_id))
    assert len(data) == 1

    assert data[0]["id"] is not None
    assert data[0]["event_id"] == event_1_id
    assert data[0]["competitor_id"] == competitor_2_id
    assert data[0]["first_name"] == "Yogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1975
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "999"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult()
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] == club_id
    assert data[0]["club"] == "OL Bundestag"
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"
    assert data[0]["result"] == result_type.PersonRaceResult()
    assert data[0]["start"] == start_type.PersonRaceStart()

    data = list(db.get_competitor(id=competitor_2_id))
    assert len(data) == 1

    assert data[0]["id"] == competitor_2_id
    assert data[0]["first_name"] == "Yogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1975
    assert data[0]["chip"] == "999"
    assert data[0]["club_id"] == club_id


def test_add_entry_result(db, event_2_id, class_2_id, club_id, entry_2_id):
    entry_id_1_result = db.add_entry_result(
        event_id=event_2_id,
        chip="4455",
        result=result_type.PersonRaceResult(status=ResultStatus.DID_NOT_START),
        start_time=S3,
    )
    entry_id_2_result = db.add_entry_result(
        event_id=event_2_id,
        chip="2289",
        result=result_type.PersonRaceResult(status=ResultStatus.MISSING_PUNCH),
        start_time=S2,
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 3

    assert data[0]["id"] == entry_id_2_result
    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] is None
    assert data[0]["last_name"] is None
    assert data[0]["gender"] is None
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "2289"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.MISSING_PUNCH
    )
    assert data[0]["start"] == start_type.PersonRaceStart(start_time=S2)
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] is None
    assert data[0]["class_"] is None

    assert data[1]["id"] == entry_id_1_result
    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] is None
    assert data[1]["last_name"] is None
    assert data[1]["gender"] is None
    assert data[1]["year"] is None
    assert data[1]["not_competing"] is False
    assert data[1]["chip"] == "4455"
    assert data[1]["fields"] == {}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S3)
    assert data[1]["club_id"] is None
    assert data[1]["club"] is None
    assert data[1]["class_id"] is None
    assert data[1]["class_"] is None

    assert data[2]["id"] == entry_2_id
    assert data[2]["event_id"] == event_2_id
    assert data[2]["first_name"] == "Angela"
    assert data[2]["last_name"] == "Merkel"
    assert data[2]["gender"] == "F"
    assert data[2]["year"] == 1957
    assert data[2]["not_competing"] is True
    assert data[2]["chip"] == "9999999"
    assert data[2]["fields"] == {0: "x"}
    assert data[2]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.DID_NOT_START
    )
    assert data[2]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[2]["club_id"] == club_id
    assert data[2]["club"] == "OL Bundestag"
    assert data[2]["class_id"] == class_2_id
    assert data[2]["class_"] == "Class 2"


def test_import_entries_empty_db(db, event_2_id):
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "OL Bundestag",
                "not_competing": True,
                "chip": "4455",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2361
                ),
                "start": start_type.PersonRaceStart(start_time=S1),
            },
            {
                "first_name": "Jogi",
                "last_name": "Löw",
                "gender": "M",
                "year": 1960,
                "class_": "Class 2",
                "club": "",
                "chip": "1234",
                "fields": {},
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH, time=2233
                ),
            },
        ],
    )
    classes = list(db.get_classes(event_id=event_2_id))
    assert len(classes) == 2
    assert classes[0].id != classes[1].id

    assert classes[0].name == "Class 1"
    assert classes[0].short_name is None
    assert classes[0].course_id is None
    assert classes[0].course is None
    assert classes[0].params == ClassParams()

    assert classes[1].name == "Class 2"
    assert classes[1].short_name is None
    assert classes[1].course_id is None
    assert classes[1].course is None
    assert classes[1].params == ClassParams()

    clubs = list(db.get_clubs())
    assert len(clubs) == 1

    assert clubs[0].name == "OL Bundestag"

    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id is None
    assert c[0].club_name is None
    assert c[0].gender == "M"
    assert c[0].year == 1960
    assert c[0].chip == "1234"

    assert c[1].first_name == "Angela"
    assert c[1].last_name == "Merkel"
    assert c[1].club_id == clubs[0].id
    assert c[1].club_name == clubs[0].name
    assert c[1].gender == ""
    assert c[1].year is None
    assert c[1].chip == "4455"

    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] != data[1]["id"]

    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1960
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "1234"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.MISSING_PUNCH, time=2233
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == classes[1].id
    assert data[0]["class_"] == classes[1].name

    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == ""
    assert data[1]["year"] is None
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "4455"
    assert data[1]["fields"] == {}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.OK, time=2361
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[1]["club_id"] == clubs[0].id
    assert data[1]["club"] == clubs[0].name
    assert data[1]["class_id"] == classes[0].id
    assert data[1]["class_"] == classes[0].name


def test_import_entries(db, event_2_id, class_1_id, club_id):
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "OL Bundestag",
                "chip": "4455",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2361
                ),
                "start": start_type.PersonRaceStart(start_time=S1),
            },
            {
                "first_name": "Jogi",
                "last_name": "Löw",
                "gender": "M",
                "year": 1960,
                "class_": "Class 1",
                "club": "",
                "chip": "1234",
                "fields": {},
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH, time=2233
                ),
            },
        ],
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id is None
    assert c[0].club_name is None
    assert c[0].gender == "M"
    assert c[0].year == 1960
    assert c[0].chip == "1234"

    assert c[1].first_name == "Angela"
    assert c[1].last_name == "Merkel"
    assert c[1].club_id == club_id
    assert c[1].club_name == "OL Bundestag"
    assert c[1].gender == ""
    assert c[1].year is None
    assert c[1].chip == "4455"

    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] != data[1]["id"]

    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1960
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "1234"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.MISSING_PUNCH, time=2233
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"

    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == ""
    assert data[1]["year"] is None
    assert data[1]["not_competing"] is False
    assert data[1]["chip"] == "4455"
    assert data[1]["fields"] == {}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.OK, time=2361
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S1)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_1_id
    assert data[1]["class_"] == "Class 1"


def test_import_entries_already_exist(
    db, event_2_id, class_1_id, club_id, entry_2_id, entry_3_id
):
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "OL Bundestag",
                "chip": "4455",
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.OK, time=2361
                ),
                "start": start_type.PersonRaceStart(start_time=S3),
            },
            {
                "first_name": "Jogi",
                "last_name": "Löw",
                "gender": "M",
                "year": 1960,
                "class_": "Class 1",
                "club": "",
                "chip": "1234",
                "fields": {},
                "result": result_type.PersonRaceResult(
                    status=ResultStatus.MISSING_PUNCH, time=2233
                ),
            },
        ],
    )
    c = list(db.get_competitors())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].first_name == "Jogi"
    assert c[0].last_name == "Löw"
    assert c[0].club_id is None
    assert c[0].club_name is None
    assert c[0].gender == "M"
    assert c[0].year == 1960
    assert c[0].chip == ""

    assert c[1].first_name == "Angela"
    assert c[1].last_name == "Merkel"
    assert c[1].club_id == club_id
    assert c[1].club_name == "OL Bundestag"
    assert c[1].gender == "F"
    assert c[1].year == 1957
    assert c[1].chip == "9999999"

    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 2

    assert data[0]["id"] != data[1]["id"]

    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Jogi"
    assert data[0]["last_name"] == "Löw"
    assert data[0]["gender"] == "M"
    assert data[0]["year"] == 1960
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "1234"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.MISSING_PUNCH, time=2233
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"

    assert data[1]["event_id"] == event_2_id
    assert data[1]["first_name"] == "Angela"
    assert data[1]["last_name"] == "Merkel"
    assert data[1]["gender"] == "F"
    assert data[1]["year"] == 1957
    assert data[1]["not_competing"] is True
    assert data[1]["chip"] == "4455"
    assert data[1]["fields"] == {0: "x"}
    assert data[1]["result"] == result_type.PersonRaceResult(
        status=ResultStatus.OK, time=2361
    )
    assert data[1]["start"] == start_type.PersonRaceStart(start_time=S3)
    assert data[1]["club_id"] == club_id
    assert data[1]["club"] == "OL Bundestag"
    assert data[1]["class_id"] == class_1_id
    assert data[1]["class_"] == "Class 1"


def test_import_entries_with_results(db, event_2_id, class_1_id):
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "",
                "chip": "4455",
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                    ),
                    status=ResultStatus.OK,
                    time=2001,
                    split_times=[
                        result_type.SplitTime(control_code="31", status="OK", time=501),
                        result_type.SplitTime(control_code="32", status="OK", time=720),
                        result_type.SplitTime(control_code="31", status="OK", time=818),
                        result_type.SplitTime(
                            control_code="33", status="OK", time=1136
                        ),
                        result_type.SplitTime(
                            control_code="31", status="OK", time=1593
                        ),
                    ],
                ),
            },
        ],
    )
    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 1

    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Angela"
    assert data[0]["last_name"] == "Merkel"
    assert data[0]["gender"] == ""
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "4455"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        start_time=datetime.datetime(
            2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
        ),
        finish_time=datetime.datetime(
            2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
        ),
        status=ResultStatus.OK,
        time=2001,
        split_times=[
            result_type.SplitTime(control_code="31", status="OK", time=501),
            result_type.SplitTime(control_code="32", status="OK", time=720),
            result_type.SplitTime(control_code="31", status="OK", time=818),
            result_type.SplitTime(control_code="33", status="OK", time=1136),
            result_type.SplitTime(control_code="31", status="OK", time=1593),
        ],
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"


def test_import_entries_already_exist_with_results(db, event_2_id, class_1_id):
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "",
                "chip": "4455",
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 0, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=datetime.datetime(
                        2020, 2, 9, 10, 33, 21, tzinfo=timezone(timedelta(hours=1))
                    ),
                    status=ResultStatus.OK,
                    time=2001,
                    split_times=[
                        result_type.SplitTime(control_code="31", status="OK", time=501),
                        result_type.SplitTime(control_code="32", status="OK", time=720),
                        result_type.SplitTime(control_code="31", status="OK", time=818),
                        result_type.SplitTime(
                            control_code="33", status="OK", time=1136
                        ),
                        result_type.SplitTime(
                            control_code="31", status="OK", time=1593
                        ),
                    ],
                ),
            },
        ],
    )
    db.import_entries(
        event_id=event_2_id,
        entries=[
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "gender": "",
                "year": None,
                "class_": "Class 1",
                "club": "",
                "chip": "1324",
                "result": result_type.PersonRaceResult(
                    start_time=datetime.datetime(
                        2020, 2, 9, 10, 5, 0, tzinfo=timezone(timedelta(hours=1))
                    ),
                    finish_time=None,
                    status=ResultStatus.MISSING_PUNCH,
                    time=1753,
                    split_times=[
                        result_type.SplitTime(control_code="31", status="OK", time=333),
                        result_type.SplitTime(control_code="34", status="OK", time=720),
                        result_type.SplitTime(
                            control_code="31", status="Missing", time=None
                        ),
                        result_type.SplitTime(
                            control_code="33", status="OK", time=1136
                        ),
                        result_type.SplitTime(
                            control_code="31", status="OK", time=1593
                        ),
                    ],
                ),
            },
        ],
    )

    data = list(db.get_entries(event_id=event_2_id))
    assert len(data) == 1

    assert data[0]["event_id"] == event_2_id
    assert data[0]["first_name"] == "Angela"
    assert data[0]["last_name"] == "Merkel"
    assert data[0]["gender"] == ""
    assert data[0]["year"] is None
    assert data[0]["not_competing"] is False
    assert data[0]["chip"] == "1324"
    assert data[0]["fields"] == {}
    assert data[0]["result"] == result_type.PersonRaceResult(
        start_time=datetime.datetime(
            2020, 2, 9, 10, 5, 0, tzinfo=timezone(timedelta(hours=1))
        ),
        finish_time=None,
        status=ResultStatus.MISSING_PUNCH,
        time=1753,
        split_times=[
            result_type.SplitTime(control_code="31", status="OK", time=333),
            result_type.SplitTime(control_code="34", status="OK", time=720),
            result_type.SplitTime(control_code="31", status="Missing", time=None),
            result_type.SplitTime(control_code="33", status="OK", time=1136),
            result_type.SplitTime(control_code="31", status="OK", time=1593),
        ],
    )
    assert data[0]["start"] == start_type.PersonRaceStart()
    assert data[0]["club_id"] is None
    assert data[0]["club"] is None
    assert data[0]["class_id"] == class_1_id
    assert data[0]["class_"] == "Class 1"



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from typing import List
from typing import Dict
from typing import Tuple

from lxml import etree
from lxml.builder import ElementMaker
import iso8601

from ooresults.repo import result_type
from ooresults.repo.result_type import ResultStatus


schema_file = pathlib.Path(__file__).parent.parent / "schema" / "IOF.xsd"
xml_schema = etree.XMLSchema(etree.parse(str(schema_file)))


iof_namespace = "http://www.orienteering.org/datastandard/3.0"
namespaces = {None: iof_namespace}


def create_result_list(event: Dict, results: List[Dict]) -> bytes:
    E = ElementMaker(namespace=iof_namespace, nsmap=namespaces)

    RESULTLIST = E.ResultList
    EVENT = E.Event
    NAME = E.Name
    STARTTIME = E.StartTime
    DATE = E.Date
    CLASSRESULT = E.ClassResult
    CLASS = E.Class
    PERSONRESULT = E.PersonResult
    PERSON = E.Person
    FAMILY = E.Family
    GIVEN = E.Given
    BIRTHDATE = E.BirthDate
    ORGANISATION = E.Organisation
    RESULT = E.Result
    CONTROLCARD = E.ControlCard
    POSITION = E.Position
    STATUS = E.Status
    STARTTIME = E.StartTime
    FINISHTIME = E.FinishTime
    TIME = E.Time
    SPLITTIME = E.SplitTime
    CONTROLCODE = E.ControlCode

    root = RESULTLIST(iofVersion="3.0")

    e_event = EVENT()
    e_event.append(NAME(event.get("name", "")))
    if event.get("date", None) is not None:
        e_event.append(STARTTIME(DATE(event["date"].isoformat())))
    root.append(e_event)

    classes: Dict[str, List[Dict]] = {}
    for result in results:
        if result["class_"] is not None:
            if result["class_"] not in classes:
                classes[result["class_"]] = []
            classes[result["class_"]].append(result)

    for c in classes:
        cr = CLASSRESULT()
        cr.append(CLASS(NAME(c)))
        for r in classes[c]:
            pr = PERSONRESULT()
            person = PERSON(
                NAME(
                    FAMILY(r["last_name"]),
                    GIVEN(r["first_name"]),
                ),
            )
            if r.get("gender", "") != "":
                person.set("sex", r["gender"])
            if r.get("year", None) is not None:
                person.append(BIRTHDATE(str(r["year"]) + "-01-01"))
            pr.append(person)

            if r.get("club", "") != "" and r.get("club", None) is not None:
                pr.append(
                    ORGANISATION(
                        NAME(r["club"]),
                    ),
                )

            result = RESULT()
            if r["result"].start_time is not None:
                result.append(
                    STARTTIME(r["result"].start_time.isoformat(timespec="seconds"))
                )
            if r["result"].finish_time is not None:
                result.append(
                    FINISHTIME(r["result"].finish_time.isoformat(timespec="seconds"))
                )
            if r["result"].time is not None:
                result.append(TIME(str(r["result"].time)))

            if r["result"].status == ResultStatus.OK:
                if r["not_competing"]:
                    result.append(STATUS("NotCompeting"))
                else:
                    if "rank" in r:
                        result.append(POSITION(str(r["rank"])))
                    result.append(STATUS("OK"))
            else:
                result.append(
                    STATUS(
                        {
                            ResultStatus.INACTIVE: "Inactive",
                            ResultStatus.OK: "OK",
                            ResultStatus.MISSING_PUNCH: "MissingPunch",
                            ResultStatus.DID_NOT_START: "DidNotStart",
                            ResultStatus.DID_NOT_FINISH: "DidNotFinish",
                            ResultStatus.DISQUALIFIED: "Disqualified",
                            ResultStatus.OVER_TIME: "OverTime",
                            ResultStatus.FINISHED: "Finished",
                        }[r["result"].status]
                    )
                )
            for s in r["result"].split_times:
                split_time = SPLITTIME(CONTROLCODE(s.control_code))
                if s.status is not None and s.status != "OK":
                    split_time.set("status", s.status)
                if s.time is not None:
                    split_time.append(TIME(str(s.time)))
                result.append(split_time)
            if r.get("chip", "") != "":
                result.append(CONTROLCARD(r["chip"], punchingSystem="SI"))
            pr.append(result)
            cr.append(pr)
        root.append(cr)

    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    return etree.tostring(
        root, encoding="UTF-8", xml_declaration=True, pretty_print=True
    )


STATUS_MAP = {
    "OK": ResultStatus.OK,
    "Finished": ResultStatus.FINISHED,
    "MissingPunch": ResultStatus.MISSING_PUNCH,
    "Disqualified": ResultStatus.DISQUALIFIED,
    "DidNotFinish": ResultStatus.DID_NOT_FINISH,
    "Active": ResultStatus.INACTIVE,
    "Inactive": ResultStatus.INACTIVE,
    "OverTime": ResultStatus.OVER_TIME,
    "SportingWithdrawal": ResultStatus.DID_NOT_FINISH,
    "NotCompeting": ResultStatus.OK,
    "Moved": ResultStatus.INACTIVE,
    "MovedUp": ResultStatus.INACTIVE,
    "DidNotStart": ResultStatus.DID_NOT_START,
    "DidNotEnter": ResultStatus.INACTIVE,
    "Cancelled": ResultStatus.INACTIVE,
}


def parse_result_list(content: bytes) -> Tuple[Dict, List[Dict]]:
    root = etree.XML(content)
    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    if not root.tag == "{" + iof_namespace + "}ResultList":
        raise RuntimeError("Root element is " + root.tag + " but should be ResultList")

    event = {}
    event["name"] = root.find("Event/Name", namespaces=namespaces).text
    date = root.find("Event/StartTime/Date", namespaces=namespaces)
    if date is not None:
        event["date"] = iso8601.parse_date(date.text).date()

    result = []
    class_result_list = root.findall("ClassResult", namespaces=namespaces)
    for c in class_result_list:
        class_ = c.find("Class/Name", namespaces=namespaces).text
        person_result_list = c.findall("PersonResult", namespaces=namespaces)
        for pr in person_result_list:
            r = {
                "first_name": "",
                "last_name": "",
                "class_": class_,
                "club": "",
                "chip": "",
                "gender": "",
                "year": None,
                "not_competing": False,
                "result": result_type.PersonRaceResult(),
            }

            r["last_name"] = pr.find("Person/Name/Family", namespaces=namespaces).text
            r["first_name"] = pr.find("Person/Name/Given", namespaces=namespaces).text

            e_person = pr.find("Person", namespaces=namespaces)
            if e_person.get("sex") is not None:
                r["gender"] = e_person.get("sex")
            e_birthdate = pr.find("Person/BirthDate", namespaces=namespaces)
            if e_birthdate is not None:
                r["year"] = int(e_birthdate.text[0:4])

            e_organization = pr.find("Organisation/Name", namespaces=namespaces)
            if e_organization is not None:
                r["club"] = e_organization.text

            e_result = pr.find("Result", namespaces=namespaces)
            e_controlcard = e_result.find("ControlCard", namespaces=namespaces)
            if e_controlcard is not None:
                r["chip"] = e_controlcard.text

            e_status = e_result.find("Status", namespaces=namespaces)
            r["result"].status = STATUS_MAP[e_status.text]
            if e_status.text == "NotCompeting":
                r["not_competing"] = True

            e_start_time = e_result.find("StartTime", namespaces=namespaces)
            if e_start_time is not None:
                r["result"].start_time = iso8601.parse_date(e_start_time.text)
                r["result"].punched_start_time = r["result"].start_time
            e_finish_time = e_result.find("FinishTime", namespaces=namespaces)
            if e_finish_time is not None:
                r["result"].finish_time = iso8601.parse_date(e_finish_time.text)
                r["result"].punched_finish_time = r["result"].finish_time
            e_time = e_result.find("Time", namespaces=namespaces)
            if e_time is not None:
                r["result"].time = int(e_time.text)

            e_split_time_list = e_result.findall("SplitTime", namespaces=namespaces)
            for e_split_time in e_split_time_list:
                split_time = result_type.SplitTime(
                    control_code=e_split_time.find(
                        "ControlCode", namespaces=namespaces
                    ).text,
                    status=e_split_time.get("status", "OK"),
                )
                e_time = e_split_time.find("Time", namespaces=namespaces)
                if e_time is not None:
                    split_time.time = int(e_time.text)
                r["result"].split_times.append(split_time)

            result.append(r)

    return event, result



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from typing import List
from typing import Dict
from typing import Tuple

import iso8601
from lxml import etree
from lxml.builder import ElementMaker


schema_file = pathlib.Path(__file__).parent.parent / "schema" / "IOF.xsd"
xml_schema = etree.XMLSchema(etree.parse(str(schema_file)))


iof_namespace = "http://www.orienteering.org/datastandard/3.0"
namespaces = {None: iof_namespace}


def create_course_data(event: Dict, courses: List[Dict], classes: List[Dict]) -> bytes:
    E = ElementMaker(namespace=iof_namespace, nsmap=namespaces)

    COURSEDATA = E.CourseData
    EVENT = E.Event
    NAME = E.Name
    STARTTIME = E.StartTime
    DATE = E.Date
    RACECOURSEDATA = E.RaceCourseData
    COURSE = E.Course
    LENGTH = E.Length
    CLIMB = E.Climb
    COURSECONTROL = E.CourseControl
    CONTROL = E.Control
    CLASSCOURSEASSIGNMENT = E.ClassCourseAssignment
    CLASSNAME = E.ClassName
    COURSENAME = E.CourseName

    root = COURSEDATA(iofVersion="3.0")

    e_event = EVENT()
    e_event.append(NAME(event.get("name", "")))
    if event.get("date", None) is not None:
        e_event.append(STARTTIME(DATE(event["date"].isoformat())))
    root.append(e_event)

    race_course_data = RACECOURSEDATA()
    for c in courses:
        course = COURSE()
        course.append(NAME(c["name"]))

        if c.get("length", None) is not None:
            course.append(LENGTH(str(c["length"])))
        if c.get("climb", None) is not None:
            course.append(CLIMB(str(c["climb"])))

        cc = COURSECONTROL(CONTROL("S1"))
        cc.set("type", "Start")
        course.append(cc)

        for control in c["controls"]:
            cc = COURSECONTROL(CONTROL(control))
            cc.set("type", "Control")
            course.append(cc)

        cc = COURSECONTROL(CONTROL("F1"))
        cc.set("type", "Finish")
        course.append(cc)

        race_course_data.append(course)

    for c in classes:
        class_course_assignment = CLASSCOURSEASSIGNMENT()
        class_course_assignment.append(CLASSNAME(c["name"]))

        if c.get("course", None) is not None:
            class_course_assignment.append(COURSENAME(c["course"]))

        race_course_data.append(class_course_assignment)

    root.append(race_course_data)

    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    return etree.tostring(
        root, encoding="UTF-8", xml_declaration=True, pretty_print=True
    )


def parse_course_data(content: bytes) -> Tuple[Dict, List[Dict], List[Dict]]:
    # Event: Dict of {'name': str, 'date': Optional[str]}
    # Course: Dict of {'name': str, 'length': Optional[float], 'climb': Optional[float], 'controls': List[str]}
    # ClassCourseAssignment: Dict of {'class_name': str, 'course_name': Optional[str]}

    root = etree.XML(content)
    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    if not root.tag == "{" + iof_namespace + "}CourseData":
        raise RuntimeError("Root element is " + root.tag + " but should be CourseData")

    event = {}
    event["name"] = root.find("Event/Name", namespaces=namespaces).text
    date = root.find("Event/StartTime/Date", namespaces=namespaces)
    if date is not None:
        event["date"] = iso8601.parse_date(date.text).date()

    result = []
    race_course_data_list = root.findall("RaceCourseData", namespaces=namespaces)

    course_list = race_course_data_list[0].findall("Course", namespaces=namespaces)
    courses = []
    for course in course_list:
        c = {"name": "", "length": None, "climb": None, "controls": []}
        c["name"] = course.find("Name", namespaces=namespaces).text

        length = course.find("Length", namespaces=namespaces)
        if length is not None:
            c["length"] = float(length.text)
        climb = course.find("Climb", namespaces=namespaces)
        if climb is not None:
            c["climb"] = float(climb.text)

        course_control_list = course.findall("CourseControl", namespaces=namespaces)
        for course_control in course_control_list:
            control_type = course_control.get("type", default="Control")
            if control_type == "Control":
                control_list = course_control.findall("Control", namespaces=namespaces)
                c["controls"].append(control_list[0].text)

        courses.append(c)

    class_course_assignment_list = race_course_data_list[0].findall(
        "ClassCourseAssignment", namespaces=namespaces
    )
    class_course_assignments = []
    for assignment in class_course_assignment_list:
        a = {
            "class_name": "",
            "course_name": None,
        }
        a["class_name"] = assignment.find("ClassName", namespaces=namespaces).text
        course_name = assignment.find("CourseName", namespaces=namespaces)
        if course_name is not None:
            a["course_name"] = course_name.text

        class_course_assignments.append(a)

    return event, courses, class_course_assignments



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import argparse
import copy
import re
import base64
import pathlib
import sys
import sqlite3
from typing import Optional

import web
from cheroot.server import HTTPServer
from cheroot.ssl.builtin import BuiltinSSLAdapter

import ooresults.handler.competitors
import ooresults.handler.classes
import ooresults.handler.courses
import ooresults.handler.clubs
import ooresults.handler.series
import ooresults.handler.results
import ooresults.handler.entries
import ooresults.handler.events
import ooresults.handler.demo_reader
from ooresults import configuration
from ooresults.handler import model
from ooresults.handler import results
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.user import Users
from ooresults.utils.globals import t_globals
from ooresults.websocket_server.websocket_handler import WebSocketHandler
from ooresults.websocket_server.websocket_server import WebSocketServer


web.config.debug = False


### Templates
templates = pathlib.Path(__file__).resolve().parent / "templates"
render_base = web.template.render(templates, base="base", globals={"str": str})
render = web.template.render(templates, globals=t_globals)


class Root:
    def GET(self):
        events = list(model.get_events())
        for event in events:
            if event.publish:
                classes = list(model.get_classes(event_id=event.id))
                entry_list = list(model.get_entries(event_id=event.id))
                class_results = results.build_results(
                    classes, copy.deepcopy(entry_list)
                )
                columns = results.build_columns(class_results)
                results_table = render.results_table(event, class_results, columns)
                return render.root(results_table)
        else:
            return render.root(None)


class Login:
    def GET(self):
        try:
            Users.update()
        except Exception as e:
            return str(e)
        raise web.seeother("/")


class Admin:
    def GET(self):
        event = {}

        events_table = render.events_table(model.get_events())
        events_tab_content = render.events_tab_content(events_table)
        entries_table = render.entries_table(event, [])
        entries_tab_content = render.entries_tab_content(entries_table)
        classes_table = render.classes_table(event, [])
        classes_tab_content = render.classes_tab_content(classes_table)
        courses_table = render.courses_table(event, [])
        courses_tab_content = render.courses_tab_content(courses_table)
        results_table = render.results_table(
            event, ooresults.handler.results.build_results([], []), set()
        )
        results_tab_content = render.results_tab_content(results_table)
        series_table = render.series_table(0, [])
        series_tab_content = render.series_tab_content(series_table)
        competitors_table = render.competitors_table([])
        competitors_tab_content = render.competitors_tab_content(competitors_table)
        clubs_table = render.clubs_table([])
        clubs_tab_content = render.clubs_tab_content(clubs_table)
        return render_base.main(
            events_tab_content,
            entries_tab_content,
            classes_tab_content,
            courses_tab_content,
            results_tab_content,
            series_tab_content,
            competitors_tab_content,
            clubs_tab_content,
        )


class Static:
    def GET(self, filename):
        """searches for and returns a requested static file or 404s out"""
        try:
            file = pathlib.Path(__file__).parent / "static" / filename
            with open(file, "r") as f:
                return f.read()
        except:
            web.application.notfound()  # file not found


def my_processor(handler):
    if web.ctx.env.get("PATH_INFO") not in ("/",) and not web.ctx.env.get(
        "PATH_INFO"
    ).startswith("/mystatic/"):
        auth = web.ctx.env.get("HTTP_AUTHORIZATION")
        if auth is not None:
            auth = re.sub("^Basic ", "", auth)
            auth = base64.b64decode(auth.encode("ascii")).decode("utf8")
            username, password = auth.split(":")
            if not Users.check(username=username, password=password):
                web.header("WWW-Authenticate", 'Basic realm="Authentication required"')
                web.ctx.status = "401 Unauthorized"
                return render.unauthorized()
        else:
            web.header("WWW-Authenticate", 'Basic realm="Authentication required"')
            web.ctx.status = "401 Unauthorized"
            return render.unauthorized()
    result = handler()
    return result


def main() -> Optional[int]:
    logging.basicConfig(
        format="%(asctime)s %(message)s",
        filename="ooresults_server.log",
        level=logging.INFO,
    )
    logging.info("-------- ooresults.server --------")

    ### Url mappings
    urls = (
        "/",
        "Root",
        "/admin",
        "Admin",
        "/login",
        "Login",
        "/mystatic/(.*)",
        "Static",
        "/event/update",
        "ooresults.handler.events.Update",
        "/event/add",
        "ooresults.handler.events.Add",
        "/event/fill_edit_form",
        "ooresults.handler.events.FillEditForm",
        "/event/delete",
        "ooresults.handler.events.Delete",
        "/entry/update",
        "ooresults.handler.entries.Update",
        "/entry/import",
        "ooresults.handler.entries.Import",
        "/entry/export",
        "ooresults.handler.entries.Export",
        "/entry/add",
        "ooresults.handler.entries.Add",
        "/entry/fill_edit_form",
        "ooresults.handler.entries.FillEditForm",
        "/entry/fill_competitors_form",
        "ooresults.handler.entries.FillCompetitorsForm",
        "/entry/fill_result_form",
        "ooresults.handler.entries.FillResultForm",
        "/entry/delete",
        "ooresults.handler.entries.Delete",
        "/result/update",
        "ooresults.handler.results.Update",
        "/result/pdfResult",
        "ooresults.handler.results.PdfResult",
        "/result/pdfSplittimes",
        "ooresults.handler.results.PdfSplittimes",
        "/result/pdfTeamResult",
        "ooresults.handler.results.PdfTeamResult",
        "/class/update",
        "ooresults.handler.classes.Update",
        "/class/import",
        "ooresults.handler.classes.Import",
        "/class/export",
        "ooresults.handler.classes.Export",
        "/class/add",
        "ooresults.handler.classes.Add",
        "/class/fill_edit_form",
        "ooresults.handler.classes.FillEditForm",
        "/class/delete",
        "ooresults.handler.classes.Delete",
        "/course/update",
        "ooresults.handler.courses.Update",
        "/course/import",
        "ooresults.handler.courses.Import",
        "/course/export",
        "ooresults.handler.courses.Export",
        "/course/add",
        "ooresults.handler.courses.Add",
        "/course/fill_edit_form",
        "ooresults.handler.courses.FillEditForm",
        "/course/delete",
        "ooresults.handler.courses.Delete",
        "/competitor/update",
        "ooresults.handler.competitors.Update",
        "/competitor/import",
        "ooresults.handler.competitors.Import",
        "/competitor/export",
        "ooresults.handler.competitors.Export",
        "/competitor/add",
        "ooresults.handler.competitors.Add",
        "/competitor/fill_edit_form",
        "ooresults.handler.competitors.FillEditForm",
        "/competitor/delete",
        "ooresults.handler.competitors.Delete",
        "/club/update",
        "ooresults.handler.clubs.Update",
        "/club/add",
        "ooresults.handler.clubs.Add",
        "/club/fill_edit_form",
        "ooresults.handler.clubs.FillEditForm",
        "/club/delete",
        "ooresults.handler.clubs.Delete",
        "/series/update",
        "ooresults.handler.series.Update",
        "/series/fill_settings_form",
        "ooresults.handler.series.FillSettingsForm",
        "/series/settings",
        "ooresults.handler.series.Settings",
        "/series/pdfResult",
        "ooresults.handler.series.PdfResult",
        "/series/csvResult",
        "ooresults.handler.series.CsvResult",
        "/si1",
        "ooresults.websocket_server.si.Si1",
        "/si2",
        "ooresults.websocket_server.si.Si2",
    )

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--path", type=pathlib.Path, dest="main_path")
    parser.add_argument("-d", "--database", type=pathlib.Path, dest="database")
    # web.application().run() use sys.argv to determine
    # the port number of the http server
    # we remove parsed arguments from sys.argv
    args, left = parser.parse_known_args()
    sys.argv = sys.argv[:1] + left

    main_path = args.main_path
    if main_path:
        if not args.main_path.exists():
            parser.error(f"Path {str(args.main_path)} not found")
        if not args.main_path.is_dir():
            parser.error(f"Path {str(args.main_path)} is not a directory")
    else:
        main_path = pathlib.Path.home() / ".ooresults"

    database = args.database
    if database:
        if not args.database.exists():
            parser.error(f"Database {str(args.database)} not found")
    else:
        database = main_path / "ooresults.sqlite"

    try:
        config = configuration.Config(path=main_path)
    except (RuntimeError, FileNotFoundError) as e:
        print(f'Error in file {str(main_path / "config.ini")}:')
        print(f"  {str(e)}")
        return 2

    try:
        model.db = SqliteRepo(
            db=str(database),
        )
    except (RuntimeError, sqlite3.Error):
        exc_type, exc_value, _ = sys.exc_info()
        logging.error(f"{exc_type.__module__}.{exc_type.__name__}: {exc_value}")
        print(f"{exc_type.__module__}.{exc_type.__name__}: {exc_value}")
        return 2

    if config.demo_reader:
        urls += ("/demo", "ooresults.handler.demo_reader.Update")

    Users.update(path=main_path / "users.json")

    app = web.application(urls, globals())
    HTTPServer.ssl_adapter = BuiltinSSLAdapter(
        certificate=config.ssl_cert,
        private_key=config.ssl_key,
    )

    try:
        app.add_processor(my_processor)
        websocket_handler = WebSocketHandler(demo_reader=config.demo_reader)
        try:
            model.websocket_server = WebSocketServer(
                handler=websocket_handler,
                ssl_cert=config.ssl_cert,
                ssl_key=config.ssl_key,
            )
        except:
            model.websocket_server = WebSocketServer(handler=websocket_handler)
        model.websocket_server.start()
        logging.info("Webserver started")
        app.run()
    except KeyboardInterrupt:
        logging.info("Webserver stopped")



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import sys
import time

import sireader
import serial.tools.list_ports


def connect() -> sireader.SIReaderReadout:
    errors = ""
    for port in serial.tools.list_ports.grep("sportident"):
        try:
            si = sireader.SIReaderReadout(port.device)
            return si
        except (sireader.SIReaderException, sireader.SIReaderTimeout) as msg:
            errors += f"port: {port.device}: {msg}\n"
            pass
    else:
        errors = "No SI Reader found" if errors == "" else errors
    raise sireader.SIReaderException(f"No SI Reader found. Possible reasons: {errors}")


def main():
    while True:
        try:
            try:
                # connect to base station, the station is automatically detected,
                # if this does not work, give the path to the port as an argument
                # see the pyserial documentation for further information.
                si = connect()

                # set to basic protocol mode
                si.poll_sicard()
                print("Setting legacy mode ...")
                si.set_extended_protocol(False)
                print("Legacy mode set")
                break
            except (sireader.SIReaderException, sireader.SIReaderTimeout, OSError) as e:
                print(f"SIReader or OSError exception:\n{str(e)}")
                time.sleep(2)

        except KeyboardInterrupt:
            break


if __name__ == "__main__":
    sys.exit(main())



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import pickle
import sqlite3
import io
import sys


VERSION = 8


old_module_names = [
    "o_results.handler.classes",
    "o_results.handler.clubs",
    "o_results.handler.competitors",
    "o_results.handler.courses",
    "o_results.handler.demo_reader",
    "o_results.handler.entries",
    "o_results.handler.events",
    "o_results.handler.handicap",
    "o_results.handler.model",
    "o_results.handler.results",
    "o_results.handler.series",
    "o_results.handler.webapi.result",
    "o_results.pdf.pdf",
    "o_results.pdf.result",
    "o_results.pdf.series",
    "o_results.pdf.splittimes",
    "o_results.pdf.team_result",
    "o_results.plugins.imports.competitors.oc_muenchen_database",
    "o_results.plugins.imports.entries.oo_net_text_csv",
    "o_results.plugins.imports.entries.oo_net_web",
    "o_results.plugins.imports.entries.text",
    "o_results.plugins.iof_class_list",
    "o_results.plugins.iof_competitor_list",
    "o_results.plugins.iof_course_data",
    "o_results.plugins.iof_entry_list",
    "o_results.plugins.iof_result_list",
    "o_results.plugins.oe2003",
    "o_results.repo.class_params",
    "o_results.repo.mem_repo",
    "o_results.repo.repo",
    "o_results.repo.result_type",
    "o_results.repo.series_type",
    "o_results.repo.sqlite_repo",
    "o_results.repo.start_type",
    "o_results.utils.globals",
    "o_results.websocket_server.si",
    "o_results.websocket_server.websocket_handler",
    "o_results.websocket_server.websocket_server",
]


class UnpicklerMapModules(pickle.Unpickler):
    def find_class(self, module: str, name: str):
        if module in old_module_names:
            module = module.replace("o_results.", "ooresults.")

        elif name == "ZoneInfo":
            if sys.version_info.major >= 3 and sys.version_info.minor >= 9:
                if module == "backports.zoneinfo":
                    module = "zoneinfo"
            else:
                if module == "zoneinfo":
                    module = "backports.zoneinfo"

        return pickle.Unpickler.find_class(self, module, name)


def update(path: str = "ooresults.sqlite"):
    conn = sqlite3.connect(database=path)
    try:
        c = conn.cursor()

        # classes
        c.execute("SELECT params, id FROM classes")
        for params, id in list(c.fetchall()):
            sql = "UPDATE classes SET params = ? WHERE id = ?"
            x = UnpicklerMapModules(io.BytesIO(params)).load()
            c.execute(sql, (pickle.dumps(x), id))

        # courses
        c.execute("SELECT controls, id FROM courses")
        for controls, id in list(c.fetchall()):
            sql = "UPDATE courses SET controls = ? WHERE id = ?"
            x = UnpicklerMapModules(io.BytesIO(controls)).load()
            c.execute(sql, (pickle.dumps(x), id))

        # events
        c.execute("SELECT fields, id FROM events")
        for fields, id in list(c.fetchall()):
            sql = "UPDATE events SET fields = ? WHERE id = ?"
            x = UnpicklerMapModules(io.BytesIO(fields)).load()
            c.execute(sql, (pickle.dumps(x), id))

        # entries
        c.execute("SELECT result, start, fields, id FROM entries")
        for result, start, fields, id in list(c.fetchall()):
            sql = "UPDATE entries SET result = ?, start = ?, fields = ? WHERE id = ?"
            x = UnpicklerMapModules(io.BytesIO(result)).load()
            y = UnpicklerMapModules(io.BytesIO(start)).load()
            z = UnpicklerMapModules(io.BytesIO(fields)).load()
            c.execute(sql, (pickle.dumps(x), pickle.dumps(y), pickle.dumps(z), id))

        # version
        sql = "UPDATE version SET value = ?"
        c.execute(sql, [VERSION])
        conn.commit()

        # compress database
        c.execute("VACUUM")
        conn.commit()

    except:
        logging.exception("Exception")
        conn.rollback()
        raise
    finally:
        conn.close()



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import copy
import datetime
from typing import Dict

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.repo.start_type import PersonRaceStart
from ooresults.repo.result_type import SplitTime
from ooresults.repo.result_type import PersonRaceResult
from ooresults.repo.result_type import ResultStatus
from ooresults.handler import model


entry_time = datetime.datetime(2015, 1, 1, 13, 38, 59, tzinfo=datetime.timezone.utc)
s1 = datetime.datetime(2015, 1, 1, 12, 38, 59, tzinfo=datetime.timezone.utc)
c1 = datetime.datetime(2015, 1, 1, 12, 39, 1, tzinfo=datetime.timezone.utc)
c2 = datetime.datetime(2015, 1, 1, 12, 39, 3, tzinfo=datetime.timezone.utc)
c3 = datetime.datetime(2015, 1, 1, 12, 39, 5, tzinfo=datetime.timezone.utc)
f1 = datetime.datetime(2015, 1, 1, 12, 39, 7, tzinfo=datetime.timezone.utc)


def t(a: datetime, b: datetime) -> int:
    diff = b.replace(microsecond=0) - a.replace(microsecond=0)
    return int(diff.total_seconds())


@pytest.fixture
def db():
    model.db = SqliteRepo(db=":memory:")
    return model.db


@pytest.fixture
def event_id(db):
    return db.add_event(
        name="Event",
        date=datetime.date(year=2015, month=1, day=1),
        key="4711",
        publish=False,
        series=None,
    )


@pytest.fixture
def course_id(db, event_id):
    return db.add_course(
        event_id=event_id,
        name="Bahn A",
        length=4500,
        climb=90,
        controls=["101", "102", "103"],
    )


@pytest.fixture
def class_id(db, event_id, course_id):
    return db.add_class(
        event_id=event_id,
        name="Elite",
        short_name="E",
        course_id=course_id,
        params=ClassParams(),
    )


@pytest.fixture
def class_id_without_course(db, event_id):
    return db.add_class(
        event_id=event_id,
        name="Elite",
        short_name="E",
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def entry_1(db, event_id, class_id):
    id = db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Robert",
        last_name="Lewandowski",
        gender="",
        year=None,
        class_id=class_id,
        club_id=None,
        not_competing=False,
        chip="12734",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )
    entry = db.get_entry(id=id)[0]
    return copy.deepcopy(entry)


@pytest.fixture
def entry_1_without_course(db, event_id, class_id_without_course):
    id = db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Robert",
        last_name="Lewandowski",
        gender="",
        year=None,
        class_id=class_id_without_course,
        club_id=None,
        not_competing=False,
        chip="12734",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )
    entry = db.get_entry(id=id)[0]
    return copy.deepcopy(entry)


@pytest.fixture
def entry_2(db, event_id, class_id):
    id = db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Yogi",
        last_name="Löw",
        gender="N",
        year=None,
        class_id=class_id,
        club_id=None,
        not_competing=False,
        chip="7410",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )
    entry = db.get_entry(id=id)[0]
    return copy.deepcopy(entry)


@pytest.fixture
def entry_3(db, event_id, class_id):
    id = db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Angela",
        last_name="Merkel",
        gender="F",
        year=1957,
        class_id=class_id,
        club_id=None,
        not_competing=False,
        chip="12734",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )
    entry = db.get_entry(id=id)[0]
    return copy.deepcopy(entry)


@pytest.fixture
def entry_4(db, event_id, class_id):
    id = db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="Yogi",
        last_name="Löw",
        gender="N",
        year=None,
        class_id=class_id,
        club_id=None,
        not_competing=False,
        chip="7410",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )
    db.update_entry_result(
        id=id,
        chip="7410",
        start_time=None,
        result=PersonRaceResult(
            status=ResultStatus.OK,
            start_time=s1,
            finish_time=f1,
            punched_start_time=s1,
            punched_finish_time=f1,
            time=t(s1, f1),
            split_times=[
                SplitTime(
                    control_code="101", punch_time=c1, time=t(s1, c1), status="OK"
                ),
                SplitTime(
                    control_code="102", punch_time=c2, time=t(s1, c2), status="OK"
                ),
                SplitTime(
                    control_code="103", punch_time=c3, time=t(s1, c3), status="OK"
                ),
            ],
        ),
    )
    entry = db.get_entry(id=id)[0]
    return copy.deepcopy(entry)


def assert_entries_are_equal(entry_a: Dict, entry_b: Dict) -> None:
    assert entry_a["id"] == entry_b["id"]
    assert entry_a["event_id"] == entry_b["event_id"]
    assert entry_a["first_name"] == entry_b["first_name"]
    assert entry_a["last_name"] == entry_b["last_name"]
    assert entry_a["gender"] == entry_b["gender"]
    assert entry_a["year"] == entry_b["year"]
    assert entry_a["competitor_id"] == entry_b["competitor_id"]
    assert entry_a["not_competing"] == entry_b["not_competing"]
    assert entry_a["chip"] == entry_b["chip"]
    assert entry_a["fields"] == entry_b["fields"]
    assert entry_a["result"] == entry_b["result"]
    assert entry_a["start"] == entry_b["start"]
    assert entry_a["club_id"] == entry_b["club_id"]
    assert entry_a["club"] == entry_b["club"]
    assert entry_a["class_id"] == entry_b["class_id"]
    assert entry_a["class_"] == entry_b["class_"]


def test_store_cardreader_result_if_cardnumber_is_unique(
    db, event_id, entry_1, entry_2, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "7410",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "7410",
        "firstName": "Yogi",
        "lastName": "Löw",
        "club": None,
        "class": "Elite",
        "status": ResultStatus.OK,
        "time": t(s1, f1),
        "error": None,
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 3

    entry_2["result"] = PersonRaceResult(
        status=ResultStatus.OK,
        start_time=s1,
        finish_time=f1,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=t(s1, f1),
        split_times=[
            SplitTime(control_code="101", punch_time=c1, time=t(s1, c1), status="OK"),
            SplitTime(control_code="102", punch_time=c2, time=t(s1, c2), status="OK"),
            SplitTime(control_code="103", punch_time=c3, time=t(s1, c3), status="OK"),
        ],
    )

    assert_entries_are_equal(entry_a=entries[0], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_2)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_3)


def test_store_cardreader_result_if_cardnumber_is_unknown(
    db, event_id, entry_1, entry_2, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "999999",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "999999",
        "firstName": None,
        "lastName": None,
        "club": None,
        "class": None,
        "status": ResultStatus.FINISHED,
        "time": None,
        "error": "Control card unknown",
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 4

    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        start_time=s1,
        finish_time=f1,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=t(s1, f1),
        split_times=[
            SplitTime(
                control_code="101", punch_time=c1, time=t(s1, c1), status="Additional"
            ),
            SplitTime(
                control_code="102", punch_time=c2, time=t(s1, c2), status="Additional"
            ),
            SplitTime(
                control_code="103", punch_time=c3, time=t(s1, c3), status="Additional"
            ),
        ],
    )
    new_entry = {
        "id": [
            e["id"]
            for e in entries
            if e["id"] not in [entry_1["id"], entry_2["id"], entry_3["id"]]
        ][0],
        "event_id": event_id,
        "first_name": None,
        "last_name": None,
        "gender": None,
        "year": None,
        "competitor_id": None,
        "not_competing": False,
        "chip": "999999",
        "fields": {},
        "result": result,
        "start": PersonRaceStart(),
        "club_id": None,
        "club": None,
        "class_id": None,
        "class_": None,
    }

    assert_entries_are_equal(entry_a=entries[0], entry_b=new_entry)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_2)
    assert_entries_are_equal(entry_a=entries[3], entry_b=entry_3)


def test_store_cardreader_result_if_cardnumber_exist_several_times(
    db, event_id, entry_1, entry_2, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "12734",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "12734",
        "firstName": None,
        "lastName": None,
        "club": None,
        "class": None,
        "status": ResultStatus.FINISHED,
        "time": None,
        "error": "Control card assigned several times",
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 4

    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        start_time=s1,
        finish_time=f1,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=t(s1, f1),
        split_times=[
            SplitTime(
                control_code="101", punch_time=c1, time=t(s1, c1), status="Additional"
            ),
            SplitTime(
                control_code="102", punch_time=c2, time=t(s1, c2), status="Additional"
            ),
            SplitTime(
                control_code="103", punch_time=c3, time=t(s1, c3), status="Additional"
            ),
        ],
    )
    new_entry = {
        "id": [
            e["id"]
            for e in entries
            if e["id"] not in [entry_1["id"], entry_2["id"], entry_3["id"]]
        ][0],
        "event_id": event_id,
        "first_name": None,
        "last_name": None,
        "gender": None,
        "year": None,
        "competitor_id": None,
        "not_competing": False,
        "chip": "12734",
        "fields": {},
        "result": result,
        "start": PersonRaceStart(),
        "club_id": None,
        "club": None,
        "class_id": None,
        "class_": None,
    }

    assert_entries_are_equal(entry_a=entries[0], entry_b=new_entry)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_2)
    assert_entries_are_equal(entry_a=entries[3], entry_b=entry_3)


def test_store_cardreader_result_if_cardnumer_is_unique_with_same_result(
    db, event_id, entry_1, entry_4, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "7410",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "7410",
        "firstName": "Yogi",
        "lastName": "Löw",
        "club": None,
        "class": "Elite",
        "status": ResultStatus.OK,
        "time": t(s1, f1),
        "error": None,
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 3

    assert_entries_are_equal(entry_a=entries[0], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_4)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_3)


def test_store_cardreader_result_if_cardnumber_is_unique_with_another_result(
    db, event_id, entry_1, entry_4, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="109", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "7410",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "7410",
        "firstName": None,
        "lastName": None,
        "club": None,
        "class": None,
        "status": ResultStatus.FINISHED,
        "time": None,
        "error": "There is already a result",
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 4

    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        start_time=s1,
        finish_time=f1,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=t(s1, f1),
        split_times=[
            SplitTime(
                control_code="101", punch_time=c1, time=t(s1, c1), status="Additional"
            ),
            SplitTime(
                control_code="102", punch_time=c2, time=t(s1, c2), status="Additional"
            ),
            SplitTime(
                control_code="109", punch_time=c3, time=t(s1, c3), status="Additional"
            ),
        ],
    )
    new_entry = {
        "id": [
            e["id"]
            for e in entries
            if e["id"] not in [entry_1["id"], entry_4["id"], entry_3["id"]]
        ][0],
        "event_id": event_id,
        "first_name": None,
        "last_name": None,
        "gender": None,
        "year": None,
        "competitor_id": None,
        "not_competing": False,
        "chip": "7410",
        "fields": {},
        "result": result,
        "start": PersonRaceStart(),
        "club_id": None,
        "club": None,
        "class_id": None,
        "class_": None,
    }

    assert_entries_are_equal(entry_a=entries[0], entry_b=new_entry)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_4)
    assert_entries_are_equal(entry_a=entries[3], entry_b=entry_3)


def test_store_cardreader_result_use_empty_control_list_if_course_is_undefined(
    db, event_id, entry_1_without_course
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="109", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "12734",
        "result": result,
    }

    status, event, res = model.store_cardreader_result(event_key="4711", item=item)

    assert status == "cardRead"
    assert event["id"] == event_id
    assert res == {
        "entryTime": entry_time,
        "eventId": event_id,
        "controlCard": "12734",
        "firstName": "Robert",
        "lastName": "Lewandowski",
        "club": None,
        "class": "Elite",
        "status": ResultStatus.FINISHED,
        "time": t(s1, f1),
        "error": None,
    }

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 1

    entry_1_without_course["result"] = PersonRaceResult(
        status=ResultStatus.FINISHED,
        start_time=s1,
        finish_time=f1,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=t(s1, f1),
        split_times=[
            SplitTime(
                control_code="101", punch_time=c1, time=t(s1, c1), status="Additional"
            ),
            SplitTime(
                control_code="102", punch_time=c2, time=t(s1, c2), status="Additional"
            ),
            SplitTime(
                control_code="109", punch_time=c3, time=t(s1, c3), status="Additional"
            ),
        ],
    )

    assert_entries_are_equal(entry_a=entries[0], entry_b=entry_1_without_course)


def test_store_cardreader_result_exception_if_eventkey_is_unknown(
    db, event_id, entry_1, entry_2, entry_3
):
    result = PersonRaceResult(
        status=ResultStatus.FINISHED,
        punched_start_time=s1,
        punched_finish_time=f1,
        time=None,
        split_times=[
            SplitTime(control_code="101", punch_time=c1, status="Additional"),
            SplitTime(control_code="102", punch_time=c2, status="Additional"),
            SplitTime(control_code="103", punch_time=c3, status="Additional"),
        ],
    )
    item = {
        "entryType": "cardRead",
        "entryTime": entry_time,
        "controlCard": "7410",
        "result": result,
    }

    with pytest.raises(repo.EventNotFoundError, match='Event for key "4712" not found'):
        status, event, res = model.store_cardreader_result(event_key="4712", item=item)

    entries = db.get_entries(event_id=event_id)
    assert len(entries) == 3

    assert_entries_are_equal(entry_a=entries[0], entry_b=entry_1)
    assert_entries_are_equal(entry_a=entries[1], entry_b=entry_2)
    assert_entries_are_equal(entry_a=entries[2], entry_b=entry_3)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import json
import pathlib
import threading
from typing import Dict
from typing import Optional

import jsonschema


SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "username": {"type": "string"},
            "password": {"type": "string"},
        },
        "required": ["username", "password"],
        "additionalProperties": False,
    },
}


class Users:
    sema = threading.Semaphore()
    path: Optional[pathlib.Path] = None
    users: Dict[str, str] = {}

    @classmethod
    def check(cls, username: str, password: str) -> bool:
        with cls.sema:
            return username in cls.users and cls.users[username] == password

    @classmethod
    def create_default_file_if_not_exist(cls) -> None:
        users = [
            {
                "username": "admin",
                "password": "admin",
            }
        ]
        if cls.path and not cls.path.exists():
            cls.path.parent.mkdir(parents=True, exist_ok=True)
            with open(cls.path, "w") as f:
                json.dump(users, f, indent=4)

    @classmethod
    def update(cls, path: Optional[pathlib.Path] = None) -> None:
        if path:
            cls.path = path

        if cls.path:
            cls.create_default_file_if_not_exist()

            with open(cls.path) as f:
                data = json.load(f)
            jsonschema.validate(instance=data, schema=SCHEMA)
            with cls.sema:
                cls.users = {}
                for user in data:
                    cls.users[user["username"]] = user["password"]



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pathlib
from typing import List
from typing import Dict
from typing import Tuple

import iso8601
from lxml import etree
from lxml.builder import ElementMaker

from ooresults.repo import result_type


schema_file = pathlib.Path(__file__).parent.parent / "schema" / "IOF.xsd"
xml_schema = etree.XMLSchema(etree.parse(str(schema_file)))


iof_namespace = "http://www.orienteering.org/datastandard/3.0"
namespaces = {None: iof_namespace}


def create_entry_list(event: Dict, entries: List[Dict]) -> bytes:
    E = ElementMaker(namespace=iof_namespace, nsmap=namespaces)

    ENTRYLIST = E.EntryList
    EVENT = E.Event
    NAME = E.Name
    STARTTIME = E.StartTime
    DATE = E.Date
    PERSONENTRY = E.PersonEntry
    PERSON = E.Person
    FAMILY = E.Family
    GIVEN = E.Given
    BIRTHDATE = E.BirthDate
    ORGANISATION = E.Organisation
    CONTROLCARD = E.ControlCard
    CLASS = E.Class

    root = ENTRYLIST(iofVersion="3.0")

    e_event = EVENT()
    e_event.append(NAME(event.get("name", "")))
    if event.get("date", None) is not None:
        e_event.append(STARTTIME(DATE(event["date"].isoformat())))
    root.append(e_event)

    for e in entries:
        # do not export pseudo entries without name
        # (used to store not already assigned sportident results)
        if e["last_name"] is None:
            continue

        pe = PERSONENTRY()
        person = PERSON(
            NAME(
                FAMILY(e["last_name"]),
                GIVEN(e["first_name"]),
            ),
        )
        if e.get("gender", "") != "":
            person.set("sex", e["gender"])
        if e.get("year", None) is not None:
            person.append(BIRTHDATE(str(e["year"]) + "-01-01"))
        pe.append(person)

        if e.get("club", "") != "" and e.get("club", None) is not None:
            pe.append(
                ORGANISATION(
                    NAME(e["club"]),
                ),
            )

        if e.get("chip", "") != "":
            pe.append(CONTROLCARD(e["chip"], punchingSystem="SI"))
        if e.get("class_", "") != "":
            pe.append(CLASS(NAME(e["class_"])))

        root.append(pe)

    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    return etree.tostring(
        root, encoding="UTF-8", xml_declaration=True, pretty_print=True
    )


def parse_entry_list(content: bytes) -> Tuple[Dict, List[Dict]]:
    root = etree.XML(content)
    if not xml_schema.validate(root):
        raise RuntimeError(xml_schema.error_log.last_error)
    if not root.tag == "{" + iof_namespace + "}EntryList":
        raise RuntimeError("Root element is " + root.tag + " but should be ResultList")

    event = {}
    event["name"] = root.find("Event/Name", namespaces=namespaces).text
    date = root.find("Event/StartTime/Date", namespaces=namespaces)
    if date is not None:
        event["date"] = iso8601.parse_date(date.text).date()

    entries = []
    person_entry_list = root.findall("PersonEntry", namespaces=namespaces)
    for pe in person_entry_list:
        e = {
            "first_name": "",
            "last_name": "",
            "class_": "",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
            "result": result_type.PersonRaceResult(),
        }

        e["last_name"] = pe.find("Person/Name/Family", namespaces=namespaces).text
        e["first_name"] = pe.find("Person/Name/Given", namespaces=namespaces).text

        e_person = pe.find("Person", namespaces=namespaces)
        if e_person.get("sex") is not None:
            e["gender"] = e_person.get("sex")
        e_birthdate = pe.find("Person/BirthDate", namespaces=namespaces)
        if e_birthdate is not None:
            e["year"] = int(e_birthdate.text[0:4])

        e_organization = pe.find("Organisation/Name", namespaces=namespaces)
        if e_organization is not None:
            e["club"] = e_organization.text

        e_controlcard = pe.find("ControlCard", namespaces=namespaces)
        if e_controlcard is not None:
            e["chip"] = e_controlcard.text
        e_class = pe.find("Class/Name", namespaces=namespaces)
        if e_class is not None:
            e["class_"] = e_class.text

        entries.append(e)

    return event, entries



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from typing import Optional

from ooresults.repo import result_type


class Entry:
    def __init__(
        self,
        id: int = 0,
        event_id: int = 0,
        first_name: str = "",
        last_name: str = "",
        gender: str = "",
        year: Optional[int] = None,
        not_competing: bool = False,
        competitor_id: int = 0,
        class_id: int = 0,
        class_: str = "",
        club: str = "",
        club_id: Optional[int] = None,
        chip: str = "",
        result: result_type.PersonRaceResult = result_type.PersonRaceResult(),
    ):
        self.first_name = first_name
        self.last_name = last_name
        self.gender = gender
        self.year = year
        self.not_competing = not_competing
        self.competitor_id = competitor_id
        self.class_id = class_id
        self.class_ = class_
        self.club = club
        self.club_id = club_id
        self.chip = chip
        self.result = result

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class Competitor:
    def __init__(
        self, first_name="", last_name="", club_name="", gender="", year=None, chip=""
    ):
        self.first_name = first_name
        self.last_name = last_name
        self.club_name = club_name
        self.gender = gender
        self.year = year
        self.chip = chip

    def __contains__(self, key):
        return hasattr(self, key)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def get(self, key, default):
        return getattr(self, key, default)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
from enum import Enum
from typing import Optional
from typing import Dict
from typing import List

from ooresults.repo.class_params import ClassParams
from ooresults.repo import result_type
from ooresults.repo import series_type


class ClassUsedError(RuntimeError):
    pass


class CourseUsedError(RuntimeError):
    pass


class ClubUsedError(RuntimeError):
    pass


class CompetitorUsedError(RuntimeError):
    pass


class EventNotFoundError(RuntimeError):
    pass


class ConstraintError(RuntimeError):
    pass


class OperationalError(RuntimeError):
    pass


class TransactionMode(Enum):
    DEFERRED = "DEFERRED"
    IMMEDIATE = "IMMEDIATE"
    EXCLUSIVE = "EXCLUSIVE"


class Transaction:
    """Database transaction."""

    def __init__(self, db, mode: TransactionMode = TransactionMode.DEFERRED):
        self.db = db
        self.mode = mode

    def __enter__(self):
        self.db.start_transaction(mode=self.mode)
        return self

    def __exit__(self, exctype, excvalue, traceback):
        print(exctype, excvalue)
        if exctype is not None:
            self.db.rollback()
        else:
            self.db.commit()


class Repo:
    def __init__(self):
        raise NotImplementedError

    def transaction(
        self, mode: TransactionMode = TransactionMode.DEFERRED
    ) -> Transaction:
        return Transaction(db=self, mode=mode)

    def start_transaction(self, mode: str):
        raise NotImplementedError

    def commit(self):
        raise NotImplementedError

    def rollback(self):
        raise NotImplementedError

    def get_classes(self, event_id: int):
        raise NotImplementedError

    def get_class(self, id):
        raise NotImplementedError

    def add_class(
        self,
        event_id: int,
        name: str,
        short_name: Optional[str],
        course_id: Optional[int],
        params: ClassParams,
    ):
        raise NotImplementedError

    def update_class(
        self,
        id,
        name,
        short_name: Optional[str],
        course_id: Optional[int],
        params: ClassParams,
    ):
        raise NotImplementedError

    def delete_classes(self, event_id):
        raise NotImplementedError

    def delete_class(self, id):
        raise NotImplementedError

    def get_courses(self, event_id: int):
        raise NotImplementedError

    def get_course(self, id):
        raise NotImplementedError

    def add_course(
        self,
        event_id: int,
        name: str,
        length: Optional[float],
        climb: Optional[float],
        controls: List[str],
    ):
        raise NotImplementedError

    def update_course(
        self,
        id,
        name: str,
        length: Optional[float],
        climb: Optional[float],
        controls: List[str],
    ):
        raise NotImplementedError

    def delete_courses(self, event_id):
        raise NotImplementedError

    def delete_course(self, id):
        raise NotImplementedError

    def get_clubs(self):
        raise NotImplementedError

    def get_club(self, id):
        raise NotImplementedError

    def add_club(self, name: str):
        raise NotImplementedError

    def update_club(self, id, name: str):
        raise NotImplementedError

    def delete_club(self, id):
        raise NotImplementedError

    def get_competitors(self):
        raise NotImplementedError

    def get_competitor(self, id):
        raise NotImplementedError

    def add_competitor(
        self, first_name, last_name, club_id, gender, year: Optional[int], chip
    ):
        raise NotImplementedError

    def update_competitor(
        self, id, first_name, last_name, club_id, gender, year: Optional[int], chip
    ):
        raise NotImplementedError

    def delete_competitor(self, id):
        raise NotImplementedError

    def import_competitors(self, competitors):
        raise NotImplementedError

    def delete_entries(self, event_id):
        raise NotImplementedError

    def delete_entry(self, id):
        raise NotImplementedError

    def import_entries(self, event_id, results):
        raise NotImplementedError

    def get_entries(self, event_id):
        raise NotImplementedError

    def get_entry(self, id):
        raise NotImplementedError

    def add_entry(
        self,
        event_id: int,
        competitor_id: Optional[int],
        first_name: str,
        last_name: str,
        gender: str,
        year: Optional[int],
        class_id: int,
        club_id: Optional[int],
        not_competing: bool,
        chip: str,
        fields: Dict[int, str],
        status: result_type.ResultStatus,
        start_time: Optional[datetime.datetime],
    ) -> int:
        raise NotImplementedError

    def update_entry(
        self,
        id: int,
        first_name: str,
        last_name: str,
        gender: str,
        year: Optional[int],
        class_id: int,
        club_id: Optional[int],
        not_competing: bool,
        chip: str,
        fields: Dict[int, str],
        status: result_type.ResultStatus,
        start_time: Optional[datetime.datetime],
    ) -> None:
        raise NotImplementedError

    def add_entry_result(
        self,
        event_id: int,
        chip,
        start_time: Optional[datetime.datetime],
        result: result_type.PersonRaceResult,
    ) -> int:
        raise NotImplementedError

    def update_entry_result(
        self,
        id: int,
        chip,
        start_time: Optional[datetime.datetime],
        result: result_type.PersonRaceResult,
    ) -> None:
        raise NotImplementedError

    def get_events(self) -> List:
        raise NotImplementedError

    def get_event(self, id: int) -> List:
        raise NotImplementedError

    def add_event(
        self,
        name: str,
        date: datetime.date,
        key: Optional[str],
        publish: bool,
        series: Optional[str],
        fields: List[str] = [],
    ) -> int:
        raise NotImplementedError

    def update_event(
        self,
        id: int,
        name: str,
        date: datetime.date,
        key: Optional[str],
        publish: bool,
        series: Optional[str],
        fields: List[str] = [],
    ) -> None:
        raise NotImplementedError

    def delete_event(self, id: int) -> None:
        raise NotImplementedError

    def get_series_settings(self) -> series_type.Settings:
        raise NotImplementedError

    def update_series_settings(self, settings: series_type.Settings) -> None:
        raise NotImplementedError



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import logging
import pathlib

import web
from web.utils import Storage

from ooresults.handler import model
from ooresults.repo.repo import ClubUsedError
from ooresults.repo.repo import ConstraintError


templates = pathlib.Path(__file__).resolve().parent.parent / "templates"
render = web.template.render(templates, globals={"str": str})


class Update:
    def POST(self):
        """Update data"""
        return render.clubs_table(model.get_clubs())


class Add:
    def POST(self):
        """Add or edit entry"""
        data = web.input()
        print(data)
        try:
            if data.id == "":
                model.add_club(data.name)
            else:
                model.update_club(int(data.id), data.name)
        except ConstraintError as e:
            raise web.conflict(str(e))
        except:
            logging.exception("Internal server error")
            raise

        return render.clubs_table(model.get_clubs())


class Delete:
    def POST(self):
        """Delete entry"""
        data = web.input()
        print(data)
        try:
            model.delete_club(int(data.id))
            return render.clubs_table(model.get_clubs())
        except ClubUsedError:
            raise web.conflict("Club used in competitors or entries")
        except:
            logging.exception("Internal server error")
            raise


class FillEditForm:
    def POST(self):
        """Query data to fill add or edit form"""
        data = web.input()
        if data.id == "":
            club = Storage(
                {
                    "id": "",
                    "name": "",
                }
            )
        else:
            club = model.get_club(int(data.id))[0]

        return render.add_club(club)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.class_params import ClassParams
from ooresults.repo.result_type import ResultStatus


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def club_1_id(db):
    return db.add_club(
        name="Club 1",
    )


@pytest.fixture
def club_2_id(db):
    return db.add_club(
        name="Club 2",
    )


@pytest.fixture
def competitor_id(db, club_1_id):
    return db.add_competitor(
        first_name="A",
        last_name="B",
        club_id=club_1_id,
        gender="",
        year=None,
        chip="",
    )


@pytest.fixture
def event_id(db):
    return db.add_event(
        name="Event",
        date=datetime.date(year=2020, month=1, day=1),
        key=None,
        publish=False,
        series=None,
    )


@pytest.fixture
def class_id(db, event_id):
    return db.add_class(
        event_id=event_id,
        name="Class",
        short_name=None,
        course_id=None,
        params=ClassParams(),
    )


@pytest.fixture
def entry_id(db, event_id, class_id, club_1_id):
    return db.add_entry(
        event_id=event_id,
        competitor_id=None,
        first_name="A",
        last_name="B",
        gender="",
        year=None,
        class_id=class_id,
        club_id=club_1_id,
        not_competing=False,
        chip="",
        fields={},
        status=ResultStatus.INACTIVE,
        start_time=None,
    )


def test_get_clubs_after_adding_one_club(db, club_1_id):
    c = list(db.get_clubs())
    assert len(c) == 1
    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"


def test_get_clubs_after_adding_two_clubs(db, club_1_id, club_2_id):
    c = list(db.get_clubs())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"
    assert c[1].id == club_2_id
    assert c[1].name == "Club 2"


def test_get_first_added_club(db, club_1_id, club_2_id):
    c = list(db.get_club(id=club_1_id))
    assert len(c) == 1
    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"


def test_get_last_added_club(db, club_1_id, club_2_id):
    c = list(db.get_club(id=club_2_id))
    assert len(c) == 1
    assert c[0].id == club_2_id
    assert c[0].name == "Club 2"


def test_update_first_added_club(db, club_1_id, club_2_id):
    db.update_club(id=club_1_id, name="Club 3")
    c = list(db.get_clubs())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == club_2_id
    assert c[0].name == "Club 2"
    assert c[1].id == club_1_id
    assert c[1].name == "Club 3"


def test_update_last_added_club(db, club_1_id, club_2_id):
    db.update_club(id=club_2_id, name="Club 3")
    c = list(db.get_clubs())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"
    assert c[1].id == club_2_id
    assert c[1].name == "Club 3"


def test_delete_first_added_club(db, club_1_id, club_2_id):
    db.delete_club(id=club_1_id)
    c = list(db.get_clubs())
    assert len(c) == 1
    assert c[0].id == club_2_id
    assert c[0].name == "Club 2"


def test_delete_last_added_club(db, club_1_id, club_2_id):
    db.delete_club(id=club_2_id)
    c = list(db.get_clubs())
    assert len(c) == 1
    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"


def test_add_existing_name_raises_exception(db, club_1_id):
    with pytest.raises(repo.ConstraintError, match="Club already exist"):
        db.add_club(name="Club 1")


def test_change_to_existing_name_raises_exception(db, club_1_id, club_2_id):
    with pytest.raises(repo.ConstraintError, match="Club already exist"):
        db.update_club(id=club_1_id, name="Club 2")


def test_update_with_unknown_id_raises_exception(db, club_1_id):
    with pytest.raises(KeyError):
        db.update_club(id=club_1_id + 1, name="Club 2")


def test_delete_club_with_unknown_id_do_not_change_anything(db, club_1_id):
    db.delete_club(id=club_1_id + 1)
    c = list(db.get_clubs())
    assert len(c) == 1
    assert c[0].id == club_1_id
    assert c[0].name == "Club 1"


def test_delete_club_used_in_competitor_raises_exception(db, competitor_id, club_1_id):
    with pytest.raises(repo.ClubUsedError):
        db.delete_club(id=club_1_id)


def test_delete_club_used_in_entry_raises_exception(db, entry_id, club_1_id):
    with pytest.raises(repo.ClubUsedError):
        db.delete_club(id=club_1_id)



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import pytest

from ooresults.repo.sqlite_repo import SqliteRepo
from ooresults.repo.series_type import Settings


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


def test_series_settings_defaults(db):
    s = db.get_series_settings()
    assert s == Settings()


def test_series_settings_update_1(db):
    settings = Settings(
        name="Series 1",
        nr_of_best_results=4,
        mode="Proportional 1",
        maximum_points=500,
        decimal_places=3,
    )
    db.update_series_settings(settings=settings)
    s = db.get_series_settings()
    print(s.name, s.nr_of_best_results, s.mode, s.maximum_points, s.decimal_places)
    assert db.get_series_settings() == settings


def test_series_settings_update_2(db):
    settings = Settings(
        name="Series 2",
        nr_of_best_results=None,
        mode="Proportional 2",
        maximum_points=600,
        decimal_places=4,
    )
    db.update_series_settings(settings=settings)
    assert db.get_series_settings() == settings


def test_series_settings_updates(db):
    settings = Settings(
        name="Series 1",
        nr_of_best_results=4,
        mode="Proportional 1",
        maximum_points=500,
        decimal_places=3,
    )
    db.update_series_settings(settings=settings)
    settings = Settings(
        name="Series 2",
        nr_of_best_results=None,
        mode="Proportional 2",
        maximum_points=600,
        decimal_places=4,
    )
    db.update_series_settings(settings=settings)
    assert db.get_series_settings() == settings



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

import pytest

from ooresults.repo import repo
from ooresults.repo.sqlite_repo import SqliteRepo


D_2021_03_02 = datetime.date(year=2021, month=3, day=2)
D_2021_03_18 = datetime.date(year=2021, month=3, day=18)
D_2020_02_14 = datetime.date(year=2020, month=2, day=14)
D_2020_02_03 = datetime.date(year=2020, month=2, day=3)


@pytest.fixture
def db():
    return SqliteRepo(db=":memory:")


@pytest.fixture
def event_1_id(db):
    return db.add_event(
        name="XX",
        date=D_2021_03_02,
        key="4711",
        publish=False,
        series="Run 1",
        fields=[],
    )


@pytest.fixture
def event_2_id(db):
    return db.add_event(
        name="YY",
        date=D_2021_03_18,
        key=None,
        publish=True,
        series=None,
        fields=["f1", "f2"],
    )


def test_get_events_after_adding_one_event(db, event_1_id):
    c = list(db.get_events())
    assert len(c) == 1
    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []


def test_get_events_after_adding_two_events(db, event_1_id, event_2_id):
    c = list(db.get_events())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []
    assert c[1].id == event_2_id
    assert c[1].name == "YY"
    assert c[1].date == D_2021_03_18
    assert c[1].key is None
    assert c[1].publish is True
    assert c[1].series is None
    assert c[1].fields == ["f1", "f2"]


def test_get_first_added_event(db, event_1_id, event_2_id):
    c = list(db.get_event(id=event_1_id))
    assert len(c) == 1
    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []


def test_get_last_added_event(db, event_1_id, event_2_id):
    c = list(db.get_event(id=event_2_id))
    assert len(c) == 1
    assert c[0].id == event_2_id
    assert c[0].name == "YY"
    assert c[0].date == D_2021_03_18
    assert c[0].key is None
    assert c[0].publish is True
    assert c[0].series is None
    assert c[0].fields == ["f1", "f2"]


def test_update_first_added_event(db, event_1_id, event_2_id):
    db.update_event(
        id=event_1_id,
        name="ZZ",
        date=D_2020_02_14,
        key=None,
        publish=True,
        series=None,
        fields=["x"],
    )
    c = list(db.get_events())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == event_2_id
    assert c[0].name == "YY"
    assert c[0].date == D_2021_03_18
    assert c[0].key is None
    assert c[0].publish is True
    assert c[0].series is None
    assert c[0].fields == ["f1", "f2"]
    assert c[1].id == event_1_id
    assert c[1].name == "ZZ"
    assert c[1].date == D_2020_02_14
    assert c[1].key is None
    assert c[1].publish is True
    assert c[1].series is None
    assert c[1].fields == ["x"]


def test_update_last_added_event(db, event_1_id, event_2_id):
    db.update_event(
        id=event_2_id,
        name="ZZ",
        date=D_2020_02_14,
        key="0000",
        publish=False,
        series="Run 1",
        fields=["x"],
    )
    c = list(db.get_events())
    assert len(c) == 2
    assert c[0].id != c[1].id

    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []
    assert c[1].id == event_2_id
    assert c[1].name == "ZZ"
    assert c[1].date == D_2020_02_14
    assert c[1].key == "0000"
    assert c[1].publish is False
    assert c[1].series == "Run 1"
    assert c[1].fields == ["x"]


def test_delete_first_added_event(db, event_1_id, event_2_id):
    db.delete_event(id=event_1_id)
    c = list(db.get_events())
    assert len(c) == 1
    assert c[0].id == event_2_id
    assert c[0].name == "YY"
    assert c[0].date == D_2021_03_18
    assert c[0].key is None
    assert c[0].publish is True
    assert c[0].series is None
    assert c[0].fields == ["f1", "f2"]


def test_delete_last_added_event(db, event_1_id, event_2_id):
    db.delete_event(id=event_2_id)
    c = list(db.get_events())
    assert len(c) == 1
    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []


def test_add_existing_name_raises_exception(db, event_1_id):
    with pytest.raises(repo.ConstraintError, match="Event or event key already exist"):
        db.add_event(
            name="XX", date=D_2020_02_03, key=None, publish=False, series=False
        )


def test_add_existing_key_raises_exception(db, event_1_id):
    with pytest.raises(repo.ConstraintError, match="Event or event key already exist"):
        db.add_event(
            name="ZZ", date=D_2020_02_03, key="4711", publish=True, series=False
        )


def test_change_to_existing_name_raises_exception(db, event_1_id, event_2_id):
    with pytest.raises(repo.ConstraintError, match="Event or event key already exist"):
        db.update_event(
            id=event_1_id,
            name="YY",
            date=D_2020_02_03,
            key=None,
            publish=False,
            series=True,
        )


def test_change_to_existing_key_raises_exception(db, event_1_id, event_2_id):
    with pytest.raises(repo.ConstraintError, match="Event or event key already exist"):
        db.update_event(
            id=event_2_id,
            name="ZZ",
            date=D_2020_02_03,
            key="4711",
            publish=True,
            series=True,
        )


def test_update_with_unknown_id_raises_exception(db, event_1_id):
    with pytest.raises(KeyError):
        db.update_event(
            id=event_1_id + 1,
            name="YY",
            date=D_2020_02_03,
            key=None,
            publish=False,
            series=False,
        )


def test_delete_event_with_unknown_id_do_not_change_anything(db, event_1_id):
    db.delete_event(id=event_1_id + 1)
    c = list(db.get_events())
    assert len(c) == 1
    assert c[0].id == event_1_id
    assert c[0].name == "XX"
    assert c[0].date == D_2021_03_02
    assert c[0].key == "4711"
    assert c[0].publish is False
    assert c[0].series == "Run 1"
    assert c[0].fields == []



# Copyright (C) 2022 Rainer Garus
#
# This file is part of the ooresults Python package, a software to
# compute results of orienteering events.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime

from ooresults.plugins import iof_entry_list
from ooresults.repo import result_type


def test_import_entry_list_with_one_entry():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <PersonEntry>
    <Person sex="F">
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
      <BirthDate>1972-01-01</BirthDate>
    </Person>
    <Organisation>
      <Name>OC Kanzleramt</Name>
    </Organisation>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
</EntryList>
"""
    event, entries = iof_entry_list.parse_entry_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert entries == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "OC Kanzleramt",
            "chip": "1234567",
            "gender": "F",
            "year": 1972,
            "result": result_type.PersonRaceResult(),
        },
    ]


def test_export_entry_list_with_one_entry():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <PersonEntry>
    <Person sex="F">
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
      <BirthDate>1972-01-01</BirthDate>
    </Person>
    <Organisation>
      <Name>OC Kanzleramt</Name>
    </Organisation>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
</EntryList>
"""
    document = iof_entry_list.create_entry_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "club": "OC Kanzleramt",
                "chip": "1234567",
                "gender": "F",
                "year": 1972,
                "result": result_type.PersonRaceResult(),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")


def test_import_entry_list_without_person_entries():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
</EntryList>
"""
    event, entries = iof_entry_list.parse_entry_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert entries == []


def test_export_entry_list_without_person_entries():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
</EntryList>
"""
    document = iof_entry_list.create_entry_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [],
    )
    assert document == bytes(content, encoding="utf-8")


def test_import_entry_list_with_several_entries():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
    </Person>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Claudia</Given>
      </Name>
    </Person>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
    <Class>
      <Name>Bahn B - Mittel</Name>
    </Class>
  </PersonEntry>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Birgit</Given>
      </Name>
    </Person>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
</EntryList>
"""
    event, entries = iof_entry_list.parse_entry_list(bytes(content, encoding="utf-8"))
    assert event == {
        "name": "1. O-Cup 2020",
        "date": datetime.date(year=2020, month=2, day=9),
    }
    assert entries == [
        {
            "first_name": "Angela",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
            "result": result_type.PersonRaceResult(),
        },
        {
            "first_name": "Claudia",
            "last_name": "Merkel",
            "class_": "Bahn B - Mittel",
            "club": "",
            "chip": "1234567",
            "gender": "",
            "year": None,
            "result": result_type.PersonRaceResult(),
        },
        {
            "first_name": "Birgit",
            "last_name": "Merkel",
            "class_": "Bahn A - Lang",
            "club": "",
            "chip": "",
            "gender": "",
            "year": None,
            "result": result_type.PersonRaceResult(),
        },
    ]


def test_export_entry_list_with_several_entries():
    content = """\
<?xml version='1.0' encoding='UTF-8'?>
<EntryList xmlns="http://www.orienteering.org/datastandard/3.0" iofVersion="3.0">
  <Event>
    <Name>1. O-Cup 2020</Name>
    <StartTime>
      <Date>2020-02-09</Date>
    </StartTime>
  </Event>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Angela</Given>
      </Name>
    </Person>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Claudia</Given>
      </Name>
    </Person>
    <ControlCard punchingSystem="SI">1234567</ControlCard>
    <Class>
      <Name>Bahn B - Mittel</Name>
    </Class>
  </PersonEntry>
  <PersonEntry>
    <Person>
      <Name>
        <Family>Merkel</Family>
        <Given>Birgit</Given>
      </Name>
    </Person>
    <Class>
      <Name>Bahn A - Lang</Name>
    </Class>
  </PersonEntry>
</EntryList>
"""
    document = iof_entry_list.create_entry_list(
        {
            "name": "1. O-Cup 2020",
            "date": datetime.date(year=2020, month=2, day=9),
        },
        [
            {
                "first_name": "Angela",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(),
            },
            {
                "first_name": "Claudia",
                "last_name": "Merkel",
                "class_": "Bahn B - Mittel",
                "chip": "1234567",
                "result": result_type.PersonRaceResult(),
            },
            {
                "first_name": "Birgit",
                "last_name": "Merkel",
                "class_": "Bahn A - Lang",
                "result": result_type.PersonRaceResult(),
            },
        ],
    )
    assert document == bytes(content, encoding="utf-8")
